from nwrapy.nmap_db import db
from nwrapy.colours import clr
from nwrapy.utils import *
import sys


def init(args):
    print("Connecting to", args.database)
    try:
        db.read_init(args.database)
        print("Connected to", args.database)
    except Exception as e:
        print("Failed to connect to db:", e, file=sys.stderr)
    

def getter(args):
    if db.initialised:
        target=args.target
        res=db.get_all_from_host(ip=target)
        print_dict(res)

def lister(args):
    if db.initialised:
        if args.objects=='targets':
            targets=db.get_all_hosts()
            for target in targets:
                print(target)
        elif args.objects=='commands':
            pass
        elif args.objects=='subnets':
            subnets=db.get_all_subnets()
            for subnet in subnets:
                print(subnet)
        
    
def exit(args):
    sys.exit(0)



#!/usr/bin/env python3                                                        
import socket
import re

def get_ip():
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        sock.connect(('10.255.255.255', 1))
        IP = sock.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        sock.close()
    return IP

def get_subnet(cidr):
    if cidr == '16':
        patt='(?P<subnet>[\d]{1,3}\.[\d]{1,3}\.)'
        ip=get_ip()
        m=re.search(patt, ip)
        subnet=m['subnet']+'0.0/'+str(cidr)
    elif cidr == '24':
        patt='(?P<subnet>[\d]{1,3}\.[\d]{1,3}\.[\d]{1,3}\.)'
        ip=get_ip()
        m=re.search(patt, ip)
        subnet=m['subnet']+'0/'+str(cidr)
    return subnet

if __name__=='__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('--cidr', choices=['16', '24'], required=False)
    args = parser.parse_args()
    
    if args.cidr:
        print(get_subnet(args.cidr))
    else:
        print(get_ip())



import xml.etree.ElementTree as ET
import os
import sys
import glob


class Parse_exception(Exception):
    """ """


class XML_Parser:
    def __init__(self):
        pass

    def parse_progress_xml(self, xml_file):
        """
        Find all taskprogress percentages and return the last one
        """
        perc = None
        perc = self.find_taskprogress(xml_file=xml_file)[-1]

        return perc

    def merge(self, files, output_file):
        """ """
        for f in files:
            assert f.endswith(".xml"), str(f) + " does not end with .xml"

        root = ET.Element("nmaprun")

        for xml_file in files:
            xml_tree = ET.parse(xml_file)
            xml_root = xml_tree.getroot()

            if not (root):
                root = xml_root
            else:
                root.attrib["args"] += " && " + xml_root.attrib["args"]
                root.attrib["start"] += " && " + xml_root.attrib["start"]
                root.attrib["startstr"] += " && " + xml_root.attrib["startstr"]
            hosts = xml_root.findall("host")
            ahosts = root.findall("host")
            for i in range(len(hosts)):
                new_host = hosts[i]
                append_host = True
                ip = new_host.find("address").get("addr")
                for j in range(len(ahosts)):
                    aip = ahosts[j].find("address").get("addr")
                    if ip == aip:
                        append_host = False
                if append_host:
                    root.append(new_host)

        new_tree = ET.ElementTree(root)
        with open(output_file, "wb") as fd:
            new_tree.write(fd)

    def get_command(self, xml_file):
        try:
            xml_tree = ET.parse(xml_file)
            xml_root = xml_tree.getroot()
            cmd = xml_root.get("args")
            return cmd
        except ET.ParseError:
            raise Parse_exception("Could not parse " + str(xml_file))

    def get_hosts(self, xml_file):
        hosts = {}
        for host in self.find_hosts(xml_file=xml_file):
            state = host.find("status").get("state")
            if state == "up":
                ip = host.find("address").get("addr")
                try:
                    hostname = host.find("hostnames").find("hostname").get("name")
                except AttributeError:
                    hostname = "None"
                portnrs = []
                services = []
                products = []
                versions = []
                extrainfo = []
                ostypes = []
                scripts = []
                methods = []
                ports = self.find_ports(host=host)
                for port in ports:
                    state = port.find("state").get("state")
                    if state == "open":
                        portnrs.append(port.get("portid"))
                        service = port.find("service")
                        services.append(str(service.get("name")))
                        products.append(str(service.get("product")))
                        versions.append(str(service.get("version")))
                        extrainfo.append(str(service.get("extrainfo")))
                        ostypes.append(str(service.get("ostype")))
                        try:
                            scripts.append(str(port.find("script").get("output")))
                        except AttributeError:
                            scripts.append("None")
                        # methods.append(str(service.get('method')))
                hosts[ip] = {}
                if len(hostname):
                    hosts[ip]["hostname"] = hostname
                if len(portnrs):
                    hosts[ip]["ports"] = portnrs
                    hosts[ip]["services"] = services
                    hosts[ip]["products"] = products
                    hosts[ip]["versions"] = versions
                    hosts[ip]["extrainfo"] = extrainfo
                if len(ostypes):
                    hosts[ip]["ostypes"] = ostypes
                if len(scripts):
                    hosts[ip]["scripts"] = scripts
                # hosts[ip]['methods']=methods

        return hosts

    def parse_disco_xml(self, xml_file):
        live_hosts = []
        xml_tree = ET.parse(xml_file)
        xml_root = xml_tree.getroot()
        for host in self.find_hosts(xml_file=xml_file):
            state = host.find("status").get("state")
            if state == "up":
                live_hosts.append(host.find("address").get("addr"))
        return live_hosts

    def parse_disco_ports_xml(self, xml_file):
        hosts = {}
        for host in self.find_hosts(xml_file=xml_file):
            ip = host.find("address").get("addr")
            hosts[ip] = {"ports": [], "reason": [], "service": []}
            ports = self.find_ports(host=host)
            for port in ports:
                state = port.find("state").get("state")
                if state == "open":
                    hosts[ip]["ports"].append(port.get("portid"))
                    try:
                        hosts[ip]["reason"].append(port.find("state").get("reason"))
                    except AttributeError:
                        hosts[ip]["reason"].append("none")
                    try:
                        hosts[ip]["service"].append(port.find("service").get("name"))
                    except AttributeError:
                        hosts[ip]["service"].append("none")

        return self.__replace_None(hosts=hosts)

    def parse_service_ports_xml(self, xml_file):
        hosts = {}
        for host in self.find_hosts(xml_file=xml_file):
            ip = host.find("address").get("addr")
            hosts[ip] = {
                "ports": [],
                "name": [],
                "product": [],
                "version": [],
                "extrainfo": [],
                "ostype": [],
                "method": [],
                "script_id": [],
                "script_output": [],
            }
            ports = self.find_ports(host=host)
            for port in ports:
                if self.is_open(port=port):
                    hosts[ip]["ports"].append(port.get("portid"))
                    service = port.find("service")
                    script = port.find("script")
                    hosts[ip]["name"].append(service.get("name"))
                    hosts[ip]["product"].append(service.get("product"))
                    hosts[ip]["version"].append(service.get("version"))
                    hosts[ip]["extrainfo"].append(service.get("extrainfo"))
                    hosts[ip]["ostype"].append(service.get("ostype"))
                    hosts[ip]["method"].append(service.get("method"))

                    try:
                        hosts[ip]["script_id"].append(script.get("id"))
                        hosts[ip]["script_output"].append(script.get("output"))
                    except AttributeError:
                        hosts[ip]["script_id"].append("none")
                        hosts[ip]["script_output"].append("none")

        return self.__replace_None(hosts=hosts)

    def parse_os_xml(self, xml_file):
        hosts = {}
        for host in self.find_hosts(xml_file=xml_file):
            ip = host.find("address").get("addr")
            hosts[ip] = {"name": [], "accuracy": []}
            oss = self.find_os(host=host)
            for os in oss:
                hosts[ip]["name"].append(os.get("name"))
                hosts[ip]["accuracy"].append(os.get("accuracy"))

        return self.__replace_None(hosts=hosts)

    def find_taskprogress(self, xml_file):
        xml_tree = ET.parse(xml_file)
        xml_root = xml_tree.getroot()
        return xml_root.findall("taskprogress")

    def find_hosts(self, xml_file):
        try:
            xml_tree = ET.parse(xml_file)
            xml_root = xml_tree.getroot()
            return xml_root.findall("host")
        except ET.ParseError:
            raise Parse_exception("Could not parse " + str(xml_file))

    def find_ports(self, host):
        t = host.findall("ports")
        if t:
            return host.findall("ports")[0].findall("port")
        return []

    def find_os(self, host):
        t = host.findall("os")
        if t:
            return host.findall("os")[0].findall("osmatch")
        return []

    def is_open(self, port):
        state = port.find("state").get("state")
        return state == "open"

    def __replace_None(self, hosts):
        for host, it in hosts.items():
            for key, el in it.items():
                for i in range(len(el)):
                    if hosts[host][key][i] == None:
                        hosts[host][key][i] = "none"
        return hosts


xml_parser = XML_Parser()



from nwrapy.db import DB

import os

class NMAP_DB(DB):
    """
    CreepyCrawler SQLite3 database implementation class
    """

    def db_init(self, db_name):
        """
        Initialise the database, by creating/connecting to it and creating the needed tables 
        if not already existing.
        """
        self._init(db_name=db_name)
        self.tables=self.get_tables()
        self.create_nmap_tables()
        self.initialised=True

    def read_init(self, db_name):
        """
        """
        if not(db_name.endswith('.db')):
               db_name=db_name+'.db'
        if os.path.exists(db_name):
            self._init(db_name=db_name)
            self.tables=self.get_tables()
            self.initialised=True
        else:
            raise ValueError(db_name+" does not exist")
               

    def insert_subnet(self, ip, hosts_up=0):
        subnet = '.'.join(ip.split('.')[0:3])+'.0/24'

        cursor=self.select(table_name='subnets',
                           column_names=['subnet_id'],
                           where="ip_cidr='"+subnet +"'")
        exists=False
        for row in cursor:
                exists=True

        if not exists:
            self.insert(table_name='subnets',
                        column_names=['ip_cidr', 'hosts_up'],
                        values=[self.str(subnet), str(hosts_up)])

    def insert_host(self, ip, status='up', open_ports=0, reason='none', vendor='unknown',
                    hostname='none', icmp_echo="no-response", icmp_netmask="no-response",
                    icmp_timestamp="no-response"):
        subnet_id = self.get_subnet_id(ip=ip)

        if self.host_exists(ip=ip):
            pass
        else:
            self.update_hosts_up(subnet_id=subnet_id)
        
            self.insert(table_name='hosts',
                        column_names=['ip', 'status', 'open_ports',
                                      'reason', 'vendor', 'hostname',
                                      'icmp_echo', 'icmp_netmask',
                                      'icmp_timestamp',
                                      'subnet_id'],
                        values=[self.str(ip), self.str(status), str(open_ports),
                                self.str(reason), self.str(vendor), self.str(hostname),
                                self.str(icmp_echo), self.str(icmp_netmask),
                                self.str(icmp_timestamp),
                                str(subnet_id)])

    def insert_command(self, cmd, status_code, time, datetime):
        self.insert(table_name='commands',
                    column_names=['command', 'status_code', 'time', 'datetime'],
                    values=[self.str(cmd), str(status_code), str(time), self.str(datetime)])

    def insert_port(self, ip, port, reason='none', service_id='-1'):
        host_id=self.get_host_id(ip=ip)

        if self.port_host_exists(ip=ip, port=port):
            pass
        else:
            self.insert(table_name='ports',
                        column_names=['port', 'reason',
                                      'service_id', 'host_id'],
                        values=[str(port), self.str(reason),
                                str(service_id), str(host_id)])

    def insert_service(self, ip, port, name, product='none', version='none',
                       extrainfo='none', ostype='none', method='none',
                       script_id='none', script_output='none'):
        port_id=self.get_port_id(ip=ip,port=port)

        exists=False
        try:
            # if exists
            service_id=self.get_service_id_port(port_id=port_id)
            if not(str(service_id)=='-1'):
                exists=True
        except ValueError:
            # if does not exist
            exists=False
        if exists:
            pass
        else:
            self.insert(table_name='services', 
                        column_names=['name', 'product', 'version',
                                      'extrainfo', 'ostype', 'method'],
                        values=[self.str(name), self.str(product), self.str(version),
                                self.str(extrainfo), self.str(ostype), self.str(method)])
            
            service_id=self.get_service_id(name=name, product=product, version=version)
        self.update_service_port(service_id=service_id, port_id=port_id)
        self.insert_script(port_id=port_id, id_=script_id, output=script_output)

    def insert_script(self, port_id, id_, output):
        if id_=='none':
            return
        try:
            self.get_script_id(port_id=port_id, id_=id_)
            return
        except ValueError:
            self.insert(table_name='scripts',
                        column_names=['id', 'output', 'port_id'],
                        values=[self.str(id_), self.str(output), str(port_id)])

    def insert_os(self, ip, name, accuracy):
        if name=='none':
            return

        host_id=self.get_host_id(ip=ip)
        self.insert(table_name='os',
                    column_names=['name', 'accuracy', 'host_id'],
                    values=[self.str(name), str(accuracy), str(host_id)])        


    def update_host_open_ports(self, ip, open_ports):
        host_id=self.get_host_id(ip=ip)
        self.update(table_name='hosts', column_name='open_ports',value=str(open_ports),
                    where="host_id="+str(host_id))
    
    def update_hosts_up(self, subnet_id):
        hosts_up=str(int(self.get_hosts_up(subnet_id=subnet_id))+1)
        self.update(table_name='subnets',column_name='hosts_up',value=hosts_up,
                    where="subnet_id="+str(subnet_id))

    def update_service_port(self, service_id, port_id):
        port_id=str(port_id)
        self.update(table_name='ports', column_name='service_id', value=str(service_id),
                    where="port_id="+port_id)
        
                        
    def host_exists(self, ip):
        cursor=self.select(table_name='hosts',
                           column_names=['host_id'],
                           where="ip='"+ip +"'")
        exists=False
        for row in cursor:
            if row[0]:
                exists=True
                break
        return exists

    def port_host_exists(self, ip, port):
        exists=False
        host_id=self.get_host_id(ip=ip)
        cursor=self.select(table_name='ports',
                           column_names=['port_id'],
                           where="host_id="+str(host_id)+" AND port="+str(port))
        for row in cursor:
            if row[0]:
                exists=True

        return exists

    def get_script_id(self, port_id, id_):
        cursor=self.select(table_name='scripts',
                           column_names=['script_id'],
                           where="port_id="+str(port_id)+" AND id="+self.str(id_))

        script_id=None
        for row in cursor:
            if row[0]:
                script_id=row[0]

        if not(script_id):
            raise ValueError("Could not find script_id for "+str(id_))
        return script_id
    
    def get_host_id(self, ip):
        cursor=self.select(table_name='hosts',
                           column_names=['host_id'],
                           where="ip='"+ip+"'")
        host_id=None
        for row in cursor:
            if row[0]:
                host_id=row[0]

        if not(host_id):
            raise ValueError("Could not find host_id for "+ip)
        return host_id

    def get_port_id(self, ip, port):
        host_id=self.get_host_id(ip=ip)
        cursor=self.select(table_name='ports',
                           column_names=['port_id'],
                           where="host_id='"+str(host_id)+"' AND port="+str(port))
        port_id=None
        for row in cursor:
            if row[0]:
                port_id=row[0]

        if not(port_id):
            raise ValueError("Could not find port_id for "+ip+":"+str(port))
        return port_id

    def get_service_id_port(self, port_id):
        cursor=self.select(table_name='ports',
                           column_names=['service_id'],
                           where="port_id='"+str(port_id)+"'")
        service_id=None
        for row in cursor:
            if row[0]:
                service_id=row[0]

        if not(service_id):
            raise ValueError("Could not find service_id for "+str(port_id)+"")

        return service_id

    def get_service_id(self, name, product, version):
        cursor=self.select(table_name='services',
                           column_names=['service_id'],
                           where="name="+self.str(name)+" AND product="+\
                           self.str(product)+" AND version="+self.str(version))
        service_id=None
        for row in cursor:
            if row[0]:
                service_id=row[0]

        if not(service_id):
            raise ValueError("Could not find service_id for "+name+", "+product+", "+version)

        return service_id

    def assert_subnets(self, hosts):
        for host in hosts:
            self.insert_subnet(ip=host)
        
    def get_hosts_up(self, subnet_id):
        cursor=self.select(table_name='subnets',
                           column_names=['hosts_up'],
                           where="subnet_id="+str(subnet_id))
        hosts_up=0
        for row in cursor:
            hosts_up=row[0]

        return hosts_up

    def get_all_hosts(self):
        hosts=[]
        cursor=self.select(table_name='hosts',
                           column_names=['ip'])
        for row in cursor:
            hosts.append(row[0])
        return hosts

    def get_all_subnets(self):
        subnets=[]
        cursor=self.select(table_name='subnets',
                           column_names=['ip_cidr'])
        for row in cursor:
            subnets.append(row[0])
        return subnets

    def get_hosts_from_subnet(self, subnet):
        subnet_id=self.get_subnet_id(ip=subnet)
        hosts=[]
        cursor=self.select(table_name='hosts',
                           column_names=['ip'],
                           where="subnet_id="+str(subnet_id))
        for row in cursor:
            hosts.append(row[0])
        return hosts

    def get_subnet_id(self, ip):
        subnet = '.'.join(ip.split('.')[0:3])+'.0/24'

        cursor=self.select(table_name='subnets',
                           column_names=['subnet_id'],
                           where="ip_cidr='"+subnet +"'")
        
        subnet_id=None
        for row in cursor:
            subnet_id=row[0]
            break
        
        return subnet_id

    def get_port_ids_host(self, host_id):
        port_ids=[]
        cursor=self.select(table_name='ports',
                           column_names=['port_id'],
                           where="host_id="+str(host_id))
        for row in cursor:
            port_ids.append(row[0])
        return port_ids

    def get_all_from_service(self, service_id):
        cursor=self.select(table_name='services',
                           column_names=['*'],
                           where="service_id="+str(service_id))
        return cursor

    def get_os(self, host_id):
        cursor=self.select(table_name='os',
                           column_names=['*'],
                           where="host_id="+str(host_id))
        return cursor
            
    def get_all_from_port(self, port_id):
        pass

    def get_all_from_host(self, ip):
        ignore=['host_id', 'subnet_id', 'port_id', 'service_id',
                'script_id', 'port', 'os_id']

        ## HOST INFO
        host_id=self.get_host_id(ip=ip)
        cursor=self.select(table_name='hosts',
                           column_names=['*'],
                           where='ip='+self.str(ip))
        
        host_cols=self.get_column_names(table_name='hosts')
        os_cols=self.get_column_names(table_name='os')

        host_info={}
        for row in cursor:
            i=0
            for col in row:
                if not(host_cols[i] in ignore):
                    host_info[host_cols[i]]=col
                i+=1

        host_info['os']=[]
        ctr=0
        limit=3
        for row in self.get_os(host_id=host_id):
            i=0
            for col in row:
                if not(os_cols[i] in ignore):
                    host_info['os'].append({os_cols[i]:col})
                if i>limit:
                    break
                i+=1
            ctr+=1
            if ctr>=limit:
                break

        ## PORT INFO
        port_info={}
        port_cols=self.get_column_names(table_name='ports')
        port_ids=self.get_port_ids_host(host_id=host_id)
        service_ids=[]
        for port_id in port_ids:
            cursor=self.select(table_name='ports',
                               column_names=['*'],
                               where='port_id='+str(port_id))
            for row in cursor:
                port=row[1]
                port_info[port]={}
                i=0
                for col in row:
                    if not(port_cols[i] in ignore):
                        port_info[port][port_cols[i]]=col
                    if port_cols[i]=='service_id':
                        service_ids.append(col)
                        service_cols=self.get_column_names(table_name='services')
                        c=self.get_all_from_service(service_id=col)
                        for r in c:
                            j=0
                            for scol in r:
                                if not(service_cols[j] in ignore):
                                    port_info[port][service_cols[j]]=scol
                                j+=1
                    i+=1
        

        results={}
        results['host']={}
        for key, el in host_info.items():
            results['host'][key]=el
            
        results['host']['ports']=[]
        for key, el in port_info.items():
            results['host']['ports'].append({key:el})
        return results

    def get_all(self):
        results={}
        subnets = self.get_all_subnets()

        for s in subnets:
            results[s]={}
            for host in self.get_hosts_from_subnet(subnet=s):
                results[s]=self.get_all_from_host(ip=host)

        return results

    def get_column_names(self, table_name):
        cursor=self.get_columns(table_name=table_name)

        columns=[]
        for row in cursor:
            columns.append(row[1])

        return columns
                            
    def create_nmap_tables(self):
        self.create_cmd_table()
        self.create_subnet_table()
        self.create_host_table()
        self.create_services_table()
        self.create_scripts_table()
        self.create_ports_table()
        self.create_os_table()

    def create_cmd_table(self):
        """
        """
        if not('commands' in self.tables):
            self.create_table(table_name='commands',
                              column_names=['command_id', 'command',
                                            'status_code', 'time',
                                            'datetime'],
                              column_types=['INTEGER', 'char(500)',
                                            'INTEGER', 'FLOAT',
                                            'char(50)'],
                              pk_index=0, notnull=[0])
        
    def create_host_table(self):
        """
        """
        if not('hosts' in self.tables):
            self.create_table(table_name='hosts',
                              column_names=['host_id', 'ip', 'status',
                                            'open_ports', 'reason',
                                            'vendor', 'hostname',
                                            'icmp_echo','icmp_netmask',
                                            'icmp_timestamp', 'subnet_id'],
                              column_types=['INTEGER', 'char(50)', 'char(50)',
                                            'INTEGER', 'char(50)',
                                            'char(100)', 'char(100)',
                                            'char(20)', 'char(20)',
                                            'char(20)', 'INTEGER'],
                              pk_index=0, notnull=[0])

    def create_subnet_table(self):
        """
        """
        if not('subnets' in self.tables):
            self.create_table(table_name='subnets',
                              column_names=['subnet_id', 'ip_cidr',
                                            'hosts_up'],
                              column_types=['INTEGER', 'char(50)',
                                            'INTEGER'],
                              pk_index=0, notnull=[0])

    def create_services_table(self):
        """
        """
        if not('services' in self.tables):
            self.create_table(table_name='services',
                              column_names=['service_id', 'name',
                                            'product', 'version',
                                            'extrainfo', 'ostype',
                                            'method'],
                              column_types=['INTEGER', 'char(50)',
                                            'char(100)', 'char(100)',
                                            'char(100)', 'char(100)',
                                            'char(100)'],
                              pk_index=0, notnull=[0])
                        
    def create_scripts_table(self):
        """
        """
        if not('scripts' in self.tables):
            self.create_table(table_name='scripts',
                              column_names=['script_id', 'id',
                                            'output', 'port_id'],
                              column_types=['INTEGER', 'char(50)',
                                            'char(100)', 'INTEGER'],
                              pk_index=0, notnull=[0])

    def create_ports_table(self):
        """
        """
        if not('ports' in self.tables):
            self.create_table(table_name='ports',
                              column_names=['port_id', 'port',
                                            'reason', 'service_id',
                                            'host_id'],
                              column_types=['INTEGER', 'INTEGER',
                                            'char(100)', 'INTEGER',
                                            'INTEGER'],
                              pk_index=0, notnull=[0])

    def create_os_table(self):
        """
        """
        if not('os' in self.tables):
            self.create_table(table_name='os',
                              column_names=['os_id', 'name',
                                            'accuracy', 'host_id'],
                              column_types=['INTEGER', 'char(100)',
                                            'INTEGER','INTEGER'],
                              pk_index=0, notnull=[0])

    def str(self, var):
        return "'"+str(self.clean(var))+"'"
                    
db=NMAP_DB()



import csv
import os

class csv_conv():

    def __init__(self, *args, **kwargs):
        if 'output_dir' in kwargs:
            self.basedir=kwargs['output_dir']
        else:
            self.basedir=os.path.join(os.getcwd(), 'nwrapy_output')
            
        if not os.path.isdir(self.basedir):
            os.mkdir(self.basedir)
        self.csv_dir=os.path.join(self.basedir,'csv_files')
        if not os.path.isdir(self.csv_dir):
            os.mkdir(self.csv_dir)
        
        self.sum_cols=['IP', 'status', 'open_ports', 'OS', 'ports',
                       'icmp_echo', 'icmp_netmask', 'icmp_timestamp']
        self.cols=['IP', 'OS', 'port', 'service', 'product',
                   'version', 'extra',
                   'icmp_echo', 'icmp_netmask', 'icmp_timestamp']

    def list_to_csv(self, l):
        return ' '.join(l)

    def get_key_val_dict(self, d, key):
        """
        Get the value of a specified key from a dict. Works recursively
        """
        value=None
        for k, val in d.items():
            if k==key:
                return val
            if type(val)==dict:
                value=self.get_key_val_dict(val,key)
        return self.__clean(value)

    def get_key_type_dict(self, d, type_):
        """
        Get the first key of a specified type from a dict. Works recursively
        """
        key=None
        for k, val in d.items():
            if type(k)==type_:
                return k
            if type(val)==dict:
                key=self.get_key_val_dict(val,key)
        return key

    def os_row(self, os_list):
        os_l=[]
        for os_ in os_list:
            name=self.get_key_val_dict(d=os_, key='name')
            if name:
                os_l.append(name)
        os_=' / '.join(os_l)
        return os_

    def write_summary(self, hosts_dict):
        with open(os.path.join(self.csv_dir, 'summary.csv'), 'w') as f:
            writer = csv.writer(f)
            writer.writerow(self.sum_cols)
            for ip, d in hosts_dict.items():
                row=[]
                row.append(ip)
                row.append(self.get_key_val_dict(d=d, key='status'))
                row.append(self.get_key_val_dict(d=d, key='open_ports'))

                osname=self.get_key_val_dict(d=d, key='os')
                row.append(self.os_row(os_list=osname))

                ports=self.get_key_val_dict(d=d, key='ports')
                port_l=[]
                for p in ports:
                    for k,v in p.items():
                        port_l.append(str(k))
                ports=' / '.join(port_l)
                row.append(ports)
                row.append(self.get_key_val_dict(d=d, key='icmp_echo'))
                row.append(self.get_key_val_dict(d=d, key='icmp_netmask'))
                row.append(self.get_key_val_dict(d=d, key='icmp_timestamp'))
                writer.writerow(row)
                

    def write_full(self, hosts_dict):
        with open(os.path.join(self.csv_dir, 'report.csv'), 'w') as f:
            writer = csv.writer(f)
            writer.writerow(self.cols)
            for ip, d in hosts_dict.items():
                port_l=self.get_key_val_dict(d=d, key='ports')
                osname=self.get_key_val_dict(d=d, key='os')
                os_=self.os_row(os_list=osname)
                for p in port_l:
                    port_n=0
                    for k,v in p.items():
                        port_n=k
                    service=self.get_key_val_dict(d=p,key='name')
                    product=self.get_key_val_dict(d=p,key='product')
                    version=self.get_key_val_dict(d=p,key='version')
                    extra=self.get_key_val_dict(d=p,key='extrainfo')
                    row=[ip,os_,port_n,service,product,version,extra]
                    row.append(self.get_key_val_dict(d=d,
                                                     key='icmp_echo'))
                    row.append(self.get_key_val_dict(d=d,
                                                     key='icmp_netmask'))
                    row.append(self.get_key_val_dict(d=d,
                                                     key='icmp_timestamp'))
                    writer.writerow(row)
                writer.writerow([])

    def __clean(self, string):
        if type(string)==str:
            repl={',':'/'}

            for k,v in repl.items():
                string=string.replace(k,v)

        return string



from nwrapy.colours import clr

import shlex
import subprocess
import sys
import os
import time
from datetime import datetime
from threading import Thread

class Nmap_Wrapper():
    """
    """

    def __init__(self, *args, **kwargs):
        self.disco_options=''
        self.portscan_options=''
        self.probe_options=''
        self.os_options=''
        self.xml_files=[]
        self.executed={}
        
        if 'disco_options' in kwargs:
            self.disco_options=kwargs['disco_options']
            if type(self.disco_options)==list:
                self.disco_options=' '.join(self.disco_options)
        if 'portscan_options' in kwargs:
            self.portscan_options=kwargs['portscan_options']
            if type(self.portscan_options)==list:
                self.portscan_options=' '.join(self.portscan_options)
        if 'full_portscan_options' in kwargs:
            self.full_portscan_options=kwargs['full_portscan_options']
            if type(self.full_portscan_options)==list:
                self.full_portscan_options=' '.join(self.full_portscan_options)
        if 'probe_options' in kwargs:
            self.probe_options=kwargs['probe_options']
            if type(self.probe_options)==list:
                self.probe_options=' '.join(self.probe_options)
        if 'os_options' in kwargs:
            self.os_options=kwargs['os_options']
            if type(self.os_options)==list:
                self.os_options=' '.join(self.os_options)
        if 'ssl_cipher_options' in kwargs:
            self.ssl_cipher_options=kwargs['ssl_cipher_options']
            if type(self.ssl_cipher_options)==list:
                self.ssl_cipher_options=' '.join(self.ssl_cipher_options)
        if 'ssl_cert_options' in kwargs:
            self.ssl_cert_options=kwargs['ssl_cert_options']
            if type(self.ssl_cert_options)==list:
                self.ssl_cert_options=' '.join(self.ssl_cert_options)
        if 'output_dir' in kwargs:
            self.basedir=kwargs['output_dir']
            if not os.path.isdir(self.basedir):
                os.mkdir(self.basedir)
            self.xml_dir=os.path.join(self.basedir, 'xml_files')
        else:
            self.basedir=os.path.join(os.getcwd(), 'nwrapy_output')
            if not os.path.isdir(self.basedir):
                os.mkdir(self.basedir)
            self.xml_dir=os.path.join(self.basedir,'xml_files')

        if not os.path.isdir(self.xml_dir):
            os.mkdir(self.xml_dir)

        stdout, stderr = self.execute(cmd=['which', 'nmap'])
        stdout=stdout.split('\n')[0]
        if stderr:
            print(clr.RED+'[!] nmap was not found:', stderr, clr.RESET, file=sys.stderr)
            sys.exit(1)
        else:
            #print('Nmap was found: ', stdout)
            self.nmap=stdout

    def host_discovery(self, target):
        """
        """
        self.executed={}
        self.xml_files=[]
        self.send_icmp_echo(target=target)
        self.send_icmp_netmask(target=target)
        self.send_icmp_timestamp(target=target)
        self.send_tcp_syn(target=target)

    def port_scanning(self, target, full=False):
        """
        """
        self.executed={}
        self.xml_files=[]
        if full:
            self.tcp_syn_full_port_scan(target=target)
        else:
            self.tcp_syn_port_scan(target=target)


    def service_scanning(self, target_ip, ports):
        self.service_scan(target=target_ip, ports=ports)

    def os_scanning(self, target):
        self.executed={}
        self.xml_files=[]
        self.os_scan(target=target)
        
    def SSL_cipher_scanning(self, target_ip, ports):
        print('Not implemented yet')

    def SSL_cert_scanning(self, target_ip, ports):
        print('Not implemented yet')
    
    def empty_files(self):
        self.executed={}
        self.xml_files=[]
        
    def send_icmp_echo(self, target):
        self.execute_nmap(target=target,
                          cmd="{nmap} -n -sn -PE -vv {options} {target} -oX {xml_file}",
                          xml_file="icmp_echo_host_disco.xml",
                          options=self.disco_options)

    def send_icmp_netmask(self, target):
        self.execute_nmap(target=target,
                          cmd="{nmap} -n -sn -PM -vv {options} {target} -oX {xml_file}",
                          xml_file="icmp_netmask_host_disco.xml",
                          options=self.disco_options)

    def send_icmp_timestamp(self, target):
        self.execute_nmap(target=target,
                          cmd="{nmap} -n -sn -PP -vv {options} {target} -oX {xml_file}",
                          xml_file="icmp_timestamp_host_disco.xml",
                          options=self.disco_options)

    def send_tcp_syn(self, target):
        self.execute_nmap(target=target,
                          cmd="{nmap} -PS21,22,23,25,80,113,445 -PA80,113,443 -n -sn -T4 -vv {options} {target} -oX {xml_file}",
                          xml_file="tcp_syn_host_disco.xml",
                          options=self.disco_options)
        
    def tcp_syn_port_scan(self, target):
        self.execute_nmap(target=target,
                          cmd="{nmap} --top-ports 1000 -n -Pn -sS -T4 --min-parallelism 100 -min-rate 64 -vv {options} {target} -oX {xml_file}",
                          xml_file="top_1000_portscan.xml",
                          options=self.portscan_options)

    def tcp_syn_full_port_scan(self, target):
        self.execute_nmap(target=target,
                          cmd="{nmap} -p- -n -Pn -sS -T4 --min-parallelism 100 -min-rate 128 -vv {options} {target} -oX {xml_file}",
                          xml_file="full_portscan.xml",
                          options=self.full_portscan_options)

    def service_scan(self, target, ports):
        if type(ports)==list:
            ports=','.join(ports)
        self.execute_nmap(target=target,
                          cmd="{nmap} -n -Pn -sV --script banner -T4 -vv {options} {target} -oX {xml_file}",
                          xml_file="{target}_service_scan.xml",
                          options=self.probe_options+f" -p {ports}")

    def os_scan(self, target):
        self.execute_nmap(target=target,
                          cmd="{nmap} -n -Pn -O -T4 --min-parallelism 100 --min-rate 64 -vv {options} {target} -oX {xml_file}",
                          xml_file="os_scan.xml",
                          options=self.os_options)
    

    def execute_nmap(self, target, cmd, xml_file, options):
        xml_file = os.path.join(self.xml_dir, xml_file.format(target=target).replace('/','_'))
        self.xml_files.append(xml_file)
        nmap=self.nmap
        cmd=cmd.format(nmap=nmap, options=options, target=target, xml_file=xml_file)
        cmd=shlex.split(cmd)
        self.stdout,self.stderr=None,None
        t = Thread(target=self.execute, args=(cmd,))
        t.start()
        while t.is_alive():
            print('Waiting to finish', end='\r')
            time.sleep(0.001)
        t.join()
        #return self.execute(cmd=cmd)
        return self.stdout, self.stderr
        
    def execute(self, cmd: list):
        print('Executing:\n\t'+clr.FAINT_WHITE, ' '.join(cmd)+clr.RESET)
        start=time.time()
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = process.communicate()
        end=time.time()
        t="%.4f"%(end-start)
        print('Time:', t+'s      ')
        stdout = stdout.decode('utf-8')
        stderr = stderr.decode('utf-8')
        if stderr:
            print(clr.RED+str(stderr)+clr.RESET)
        self.executed[str(datetime.now())]={'cmd':' '.join(cmd),
                                            'status_code':str(process.returncode),
                                            'time':t}
        self.stdout,self.stderr=stdout,stderr
        return stdout, stderr

    def get_subnets(ips):
        subnets=[]
        for ip in ips:
            segs=ip.split('.')
            subnets.append('.'.join(segs[0:3]))
            subnets=list(dict.fromkeys(subnets))
        return subnets



import os

class Table():
    """
    Class to allow the creation of CLI tables
    """

    def table(self, headers: list, data: list, col: str='|', rw: str='-', corner: str='+'):
        """
        """
        assert len(headers)==len(data)
        col_widths=self.column_width(headers, data)
        
        term_width=os.get_terminal_size().columns
        max_width=term_width/len(headers)
        col_width=max(col_widths)
        row_div=corner+rw*(col_width-1)

        rows=['']
        for w in col_widths:
            rows[-1]+=corner+rw*(w-1)
        rows[-1]+=corner

        # Add headers
        rows.append('')
        for i,header in enumerate(headers):
            string=col+' '+header
            rem=col_widths[i]-(len(string))
            string+=' '*rem
            rows[-1]+=string
        rows[-1]+=col

        rows.append('')
        for w in col_widths:
            rows[-1]+=corner+rw*(w-1)
        rows[-1]+=corner

        # Add data
        for i in range(len(data[0])):
            rows.append('')
            for j,da in enumerate(data):
                string=col+' '+da[i]
                rem=col_widths[j]-(len(string))
                string+=' '*rem
                rows[-1]+=string
            rows[-1]+=col

        rows.append('')
        for w in col_widths:
            rows[-1]+=corner+rw*(w-1)
        rows[-1]+=corner
        
        return rows
        
    def column_width(self, headers: list, data: list):
        """
        
        """
        lengths=[]

        for i in range(len(headers)):
            l=[len(headers[i])+3]
            for da in data[i]:
                l.append(len(da)+3)

            m=max(l)
            lengths.append(m)

        assert len(headers)==len(lengths)
        return lengths



from nwrapy.colours import clr

def print_dict(d, level=0, div="-", start='|'):
    if type(d)==dict:
        for k,e in d.items():
            if type(k)==int:
                print(clr.BOLD_WHITE+start+level*div, k, clr.RESET)
            else:
                if level==0:
                    print('#'*10)
                    print(level*div+clr.GREEN+str(k), clr.RESET)
                elif level==1:
                    print('-'*10)
                    print(start+level*div+clr.FAINT_WHITE, k, clr.RESET)
                    print('-'*10)
                else:
                    print(start+level*div+clr.FAINT_WHITE, k, clr.RESET)
            print_dict(e, level+1)
    elif type(d)==list:
        for item in d:
            print_dict(item, level+1)
    else:
        print(start+(level)*div, d, clr.RESET)



import sqlite3
import sys
from threading import Lock

class DB():

    def __init__(self):
        """
        """
        pass

    def _init(self, db_name):
        self._lock=Lock()
        self.db_name=self.__init_name(name=db_name)
        try:
            self.conn=sqlite3.connect(self.db_name)
        except Exception as e:
            print(e, file=sys.stderr)
            sys.exit(1)
            
    def __init_name(self, name):
        """
        Initialise the database name
        """
        if not(type(name)==str):
            name=str(name)
        if not(name.endswith('.db')):
            name+='.db'
            
        return name
        
    def query(self, query):
        with self._lock:
            try:
                cursor = self.conn.execute(query)
            except Exception as e:
                print('Could not execute query:', query, file=sys.stderr)
                print('Error:', e, file=sys.stderr)
                sys.exit(1)

        return cursor

    def create_table(self, table_name: str, column_names: list,
                     column_types: list, pk_index:int=0,
                     notnull:list=[0]):
        if not(len(column_types)==len(column_names)):
            raise ValueError("Different number of column types and column names")
        
        query='CREATE TABLE '+table_name+' ('

        clmns=[]
        for index in range(len(column_names)):
            line=column_names[index]+' '+column_types[index]
            if index==pk_index:
                line+=' PRIMARY KEY AUTOINCREMENT'
            if index in notnull:
                line+=' NOT NULL'
            #if not(column_names[index]==column_names[-1]):
            #    line+=','
            clmns.append(line)

        clmns=', '.join(clmns)
        query+=clmns+');'
        self.query(query=query)

    def insert(self, table_name: str, column_names: list, values: list):
        if not(len(column_names)==len(values)):
            raise ValueError("Different number of column names and values")

        query='INSERT INTO '+table_name+' ('+', '.join(column_names)+')' \
            +' VALUES ('+', '.join(values)+');'
        self.query(query=query)
        self.conn.commit()

    def update(self, table_name: str, column_name: str, value: str, where: str):
        query='UPDATE '+table_name+' SET '+column_name+'='+value+' WHERE '+where
        self.query(query=query)
        self.conn.commit()

    def select(self, table_name:str, column_names:list, where:str=None):
        query='SELECT '+','.join(column_names)+' FROM '+table_name
        if where:
            query+=' WHERE '+where

        cursor=self.query(query=query)
        return cursor

    def show_tables(self):
        query='''SELECT name FROM sqlite_schema WHERE type='table' AND name NOT LIKE 'sqlite_%';'''
        rows=self.query(query=query)
        for row in rows:
            for name in row:
                print(name)
                rs = self.get_columns(table_name=name)
                for r in rs:
                    print(r)

    def get_tables(self):
        query='''SELECT name FROM sqlite_schema WHERE type='table' AND name NOT LIKE 'sqlite_%';'''
        rows=self.query(query=query)
        names=[]
        for row in rows:
            for name in row:
                names.append(name)
        return names

    def get_columns(self, table_name):
        query="PRAGMA table_info('"+table_name+"');"
        res=self.query(query=query)
        return res

    def clean(self, string):
        repl={"'":" "}
        for k,v in repl.items():
            string=string.replace(k,v)
        return string

if __name__=='__main__':
    db = DB(db_name='test.db')
    
    db.create_table(table_name='targets',
                    column_names=['target_id', 'domain', 'protocol', 'description'],
                    column_types=['int', 'char(50)', 'char(50)', 'char(100)'],
                    pk_index=0, notnull=[0])
    db.create_table(table_name='links_internal',
                    column_names=['url', 'origin', 'status', 'target_id'],
                    column_types=['char(400)', 'char(400)', 'int', 'int'])
    db.create_table(table_name='links_external',
                    column_names=['url', 'origin', 'status', 'target_id'],
                    column_types=['char(400)', 'char(400)', 'int', 'int'])
    db.create_table(table_name='subdomains',
                    column_names=['domain', 'origin', 'target_id'],
                    column_types=['char(400)', 'char(400)', 'int'])
    db.show_tables()



#!/usr/bin/env python3
from nwrapy.nmap_db import db
from nwrapy.nmap_wrapper import Nmap_Wrapper
from nwrapy.xml_parser import xml_parser, Parse_exception
from nwrapy.csv_conv import csv_conv
from nwrapy.colours import clr
from nwrapy.interactive import *
from nwrapy.read import *
from nwrapy.get_ip import *
from nwrapy.table import Table

import re
import sys
import os
import argparse
import subprocess
import time
import shlex
from threading import Thread, active_count

####################################################################
####################################################################
# CONFIGURATION
## Not required to do anything about, but you are able to change the
## default behaviour here.
### Exclude ports with specific probes. The default beheviour tries to prevent ISPs from flagging the TerminalServer probes
EXCLUDE_PORTS = [
    "T:9100-9107",
    "T:515",
    "T:1028",
    "T:1068",
    "T:1503",
    "T:1720",
    "T:1935",
    "T:2040",
    "T:3388",
]

DISCO_OPTIONS = [
    "-g 53",  # Source port 53 (20 is FTP-DATA, also an option)
    "--randomize-hosts",
    "--data-length 56",  # 56 simulates Linux ping while 32 simulates Windows
]

PORTSCAN_OPTIONS = [
    "-g 53",  # Source port 53 (20 is FTP-DATA, also an option)
    "--randomize-hosts",
    "--data-length 56",  # 56 simulates Linux ping while 32 simulates Windows
]

FULL_PORTSCAN_OPTIONS = [
    "-g 53",  # Source port 53 (20 is FTP-DATA, also an option)
    "--randomize-hosts",
    "--data-length 56",  # 56 simulates Linux ping while 32 simulates Windows
]

PROBE_OPTIONS = [
    "-g 53",  # Source port 53 (20 is FTP-DATA, also an option)
    "--version-intensity {INTENSITY}",
    "--exclude-ports " + ",".join(EXCLUDE_PORTS),
]

OS_OPTIONS = ["-g 53"]  # Source port 53 (20 is FTP-DATA, also an option)

SSL_CIPHER_OPTIONS = ["-g 53"]  # Source port 53 (20 is FTP-DATA, also an option)

SSL_CERT_OPTIONS = ["-g 53"]  # Source port 53 (20 is FTP-DATA, also an option)

### Colours
INIT = "[" + clr.BOLD_WHITE + "***" + clr.RESET + "]"
START = "[" + clr.BOLD_WHITE + "*" + clr.RESET + "]"
DONE = "[" + clr.GREY + "#" + clr.RESET + "]"
FIN = "[" + clr.BOLD_WHITE + "***" + clr.RESET + "]"
PRMPT = clr.BOLD_WHITE + "[$]: " + clr.RESET
####################################################################
####################################################################


class ArgParser(argparse.ArgumentParser):
    """
    Argument Parser that does not exit on error. To be used in interactive mode
    """

    def error(self, message):
        clr.RED
        self.print_help(sys.stderr)


def is_root():
    if os.geteuid() == 0:
        return True
    else:
        return False


def rm_duplicates(l: list):
    """
    Removes duplicates from a list
    """
    l = list(dict.fromkeys(l))
    return l


def convert_to_nmap_target(hosts: list):
    """
    Converts a list to a string, separated by spaces. To be used by nmap.

    Parameters
    ----------
    hosts: list
        The list to filter

    Returns
    -------
    .list:
        The list without duplicates
    """
    hosts = rm_duplicates(hosts)
    return " ".join(hosts)


def autoscan(args):
    """
    Scans the default network range, where the netmask can be CIDR /16 or /24

    Parameters
    ----------
    args:
        The arguments, as parsed by argparse
    """
    args.target = get_subnet(args.cidr)
    print(START, "AUTOSCANNING", args.target)
    scan(args)


def scan(args):
    if not (is_root()):
        print(clr.RED + "[!] Please run this as root" + clr.RESET, file=sys.stderr)
        sys.exit(1)

    print(START, "Initialising")
    if not args.database.endswith(".db"):
        args.database += ".db"

    db.db_init(db_name=args.database)

    if args.version_intensity:
        INTENSITY = args.version_intensity
    else:
        INTENSITY = 6
    for i in range(len(PROBE_OPTIONS)):
        if "INTENSITY" in PROBE_OPTIONS[i]:
            PROBE_OPTIONS[i] = PROBE_OPTIONS[i].format(INTENSITY=INTENSITY)

    if args.output:
        nwrap = Nmap_Wrapper(
            output_dir=args.output,
            disco_options=DISCO_OPTIONS,
            portscan_options=PORTSCAN_OPTIONS,
            full_portscan_options=FULL_PORTSCAN_OPTIONS,
            probe_options=PROBE_OPTIONS,
            os_options=OS_OPTIONS,
            ssl_cipher_options=SSL_CIPHER_OPTIONS,
            ssl_cert_options=SSL_CERT_OPTIONS,
        )
    else:
        nwrap = Nmap_Wrapper(
            disco_options=DISCO_OPTIONS,
            portscan_options=PORTSCAN_OPTIONS,
            full_portscan_options=FULL_PORTSCAN_OPTIONS,
            probe_options=PROBE_OPTIONS,
            os_options=OS_OPTIONS,
            ssl_cipher_options=SSL_CIPHER_OPTIONS,
            ssl_cert_options=SSL_CERT_OPTIONS,
        )

    max_threads = 1
    if args.threads:
        max_threads = args.threads

    print(DONE, "Initialisation done")

    #######################################################
    print(START, "STAGE 1: Host discovery for", args.target)
    args.target = convert_to_nmap_target(args.target.split(" "))
    nwrap.host_discovery(target=args.target)
    for dt, item in nwrap.executed.items():
        db.insert_command(
            cmd=item["cmd"],
            status_code=item["status_code"],
            time=item["time"],
            datetime=dt,
        )

    hosts = []
    for f in nwrap.xml_files:
        hosts += xml_parser.parse_disco_xml(xml_file=f)
    hosts = rm_duplicates(hosts)
    db.assert_subnets(hosts=hosts)
    for host in hosts:
        icmp_echo, icmp_netmask, icmp_timestamp = (
            "no-response",
            "no-response",
            "no-response",
        )
        for f in nwrap.xml_files:
            if host in xml_parser.parse_disco_xml(xml_file=f):
                if "icmp_echo" in f:
                    icmp_echo = "echo-reply"
                elif "icmp_netmask" in f:
                    icmp_netmask = "netmask-reply"
                elif "icmp_timestamp" in f:
                    icmp_timestamp = "timestamp-reply"

        db.insert_host(
            ip=host,
            icmp_echo=icmp_echo,
            icmp_netmask=icmp_netmask,
            icmp_timestamp=icmp_timestamp,
        )
    print(DONE, "STAGE 1 finished: found", len(rm_duplicates(hosts)), "hosts up")

    #######################################################
    targets = convert_to_nmap_target(hosts=hosts)
    print(
        START,
        "STAGE 2: Port Scanning for:\n\t"
        + clr.BOLD_WHITE
        + "\n\t".join(hosts)
        + clr.RESET,
    )
    nwrap.port_scanning(target=targets, full=False)

    for dt, item in nwrap.executed.items():
        db.insert_command(
            cmd=item["cmd"],
            status_code=item["status_code"],
            time=item["time"],
            datetime=dt,
        )
    results = {}
    for f in nwrap.xml_files:
        for key, el in xml_parser.parse_disco_ports_xml(xml_file=f).items():
            results[key] = el

    nr_ports = 0
    for ip, it in results.items():
        # update open ports
        db.update_host_open_ports(ip=ip, open_ports=len(it["ports"]))
        for i in range(len(it["ports"])):
            db.insert_port(ip=ip, port=it["ports"][i], reason=it["reason"][i])
            nr_ports += 1

    print(DONE, "STAGE 2 finished")

    #######################################################
    print(
        START,
        "STAGE 3: Service Scanning",
        nr_ports,
        "ports total, of",
        len(hosts),
        "hosts",
    )

    ctr = 1
    threads = []
    nwrap.empty_files()
    for ip, it in results.items():
        print(clr.BOLD_WHITE + "Host", str(ctr) + "/" + str(len(hosts)), clr.RESET)
        ctr += 1
        if it["ports"]:
            # threads.append(Thread(target=nwrap.service_scanning, args=(ip,it['ports'])))
            # threads[-1].start()
            # while active_count() >= max_threads:
            #    print('Waiting for a thread to finish', str(active_count())+'/'+str(max_threads),
            #          end='\r')
            #    time.sleep(0.1)
            nwrap.service_scanning(target_ip=ip, ports=it["ports"])
        else:
            print("No open ports discovered")

    for index, thread in enumerate(threads):
        print("Joining thread", i, end="\r")
        thread.join()
    print("")
    print(active_count())

    for f in nwrap.xml_files:
        for ip, it in xml_parser.parse_service_ports_xml(xml_file=f).items():
            for i in range(len(it["ports"])):
                db.insert_service(
                    ip=ip,
                    port=it["ports"][i],
                    name=it["name"][i],
                    product=it["product"][i],
                    version=it["version"][i],
                    extrainfo=it["extrainfo"][i],
                    ostype=it["ostype"][i],
                    method=it["method"][i],
                    script_id=it["script_id"][i],
                    script_output=it["script_output"][i],
                )

    print(DONE, "STAGE 3 finished")

    # -----------------------------------------------------#
    print(START, "Writing intermediary results")
    if args.output:
        csv = csv_conv(output_dir=args.output)
    else:
        csv = csv_conv()

    ihosts = db.get_all_hosts()
    iresults = {}
    for host in ihosts:
        iresults[host] = db.get_all_from_host(ip=host)
    csv.write_summary(hosts_dict=iresults)
    csv.write_full(hosts_dict=iresults)
    print(DONE, "Wrote intermediary results")
    # -----------------------------------------------------#

    #######################################################
    print(START, "STAGE 4: OS Scanning", len(hosts), "hosts")
    targets = convert_to_nmap_target(hosts=hosts)
    nwrap.os_scanning(target=targets)

    for f in nwrap.xml_files:
        for ip, it in xml_parser.parse_os_xml(xml_file=f).items():
            for i in range(len(it["name"])):
                db.insert_os(ip=ip, name=it["name"][i], accuracy=it["accuracy"][i])

    print(DONE, "STAGE 4 finished")

    #######################################################
    print(START, "STAGE 5: SSL Ciphers scanning on", len(hosts), "hosts")
    ctr = 1
    nwrap.empty_files()
    for ip, it in results.items():
        print(clr.BOLD_WHITE + "Host", str(ctr) + "/" + str(len(hosts)), clr.RESET)
        ctr += 1
        if it["ports"]:
            nwrap.SSL_cipher_scanning(target_ip=host, ports=it["ports"])
        else:
            print("No open ports detected")
    print(DONE, "STAGE 5 finished")

    #######################################################
    print(START, "STAGE 6: SSL Certs scanning on", len(hosts), "hosts")
    ctr = 1
    nwrap.empty_files()
    for ip, it in results.items():
        print(clr.BOLD_WHITE + "Host", str(ctr) + "/" + str(len(hosts)), clr.RESET)
        ctr += 1
        if it["ports"]:
            nwrap.SSL_cipher_scanning(target_ip=host, ports=it["ports"])
        else:
            print("No open ports detected")
    print(DONE, "STAGE 6 finished")

    if not (args.no_full_portscan):
        #######################################################
        print(START, "STAGE 7: Full port scan on", len(hosts), "hosts")
        nwrap.port_scanning(target=targets, full=True)

        for dt, item in nwrap.executed.items():
            db.insert_command(
                cmd=item["cmd"],
                status_code=item["status_code"],
                time=item["time"],
                datetime=dt,
            )
        nresults = {}
        for f in nwrap.xml_files:
            for key, el in xml_parser.parse_disco_ports_xml(xml_file=f).items():
                nresults[key] = el

        nr_ports = 0
        for ip, it in nresults.items():
            # update open ports
            db.update_host_open_ports(ip=ip, open_ports=len(it["ports"]))
            for i in range(len(it["ports"])):
                db.insert_port(ip=ip, port=it["ports"][i], reason=it["reason"][i])
                nr_ports += 1
        print(DONE, "STAGE 7 finished")

    print(START, "Writing CSV report")

    hosts = db.get_all_hosts()
    results = {}
    for host in hosts:
        results[host] = db.get_all_from_host(ip=host)
    csv.write_summary(hosts_dict=results)
    csv.write_full(hosts_dict=results)

    print(FIN, "Finished")


def read(args):
    if args.service:
        pass
    elif args.target:
        database = args.database
        read_target(target=args.target, database=database)

    print(FIN, "Finished")

def xml(args):
    xml_file = args.file
    print(clr.BOLD + "Reading from", xml_file + clr.RESET)
    if not (os.path.exists(xml_file)):
        print(clr.BOLD + clr.RED + '"' + xml_file + '"', "does not exist" + clr.RESET)
        sys.exit(1)

    """
    with open(xml_file, "r") as fd:
        lines = fd.readlines()
    for line in lines:
        print(line.split('\n')[0])
    """
    print()
    print(clr.BOLD + "Executed command:" + clr.RESET)
    try:
        cmd = xml_parser.get_command(xml_file)
    except Parse_exception as e:
        print(clr.RED + clr.BOLD + "Failed to parse", xml_file + clr.RESET)
        print("Reason:", e)
        sys.exit(1)

    print(clr.GREY + cmd + clr.RESET + "\n")

    hosts = xml_parser.get_hosts(xml_file)
    table = Table()
    for h, d in hosts.items():
        if (
            (not (args.target) or (h == args.target))
            and (
                not (args.service)
                or ("services" in d and args.service in d["services"])
            )
            and (not (args.port) or ("ports" in d and args.port in d["ports"]))
            and (
                not (args.os_type) or ("ostypes" in d and args.os_type in d["ostypes"])
            )
            and (
                not (args.product)
                or ("products" in d and args.product in d["products"])
            )
        ):
            print(clr.BOLD + "IP: " + h + clr.RESET)
            data = []
            headers = []
            for k, v in d.items():
                if type(v) == list:
                    headers.append(k)
                    data.append(v)
                else:
                    print(clr.DIM + k + ":", v + clr.RESET)

            if len(headers):
                for row in table.table(headers, data):
                    print(clr.DIM + row + clr.RESET)

    print(FIN, "Finished")

def merge(args):
    files = args.files
    if args.new_file in files:
        print(
            clr.RED + clr.BOLD + "File",
            args.new_file,
            "is part of the merge already" + clr.RESET,
        )
        sys.exit(1)
    try:
        xml_parser.merge(files, args.new_file)
        print("Wrote output to", args.new_file)
        print(FIN, "Finished")
    except Exception as e:
        print(clr.RED + clr.BOLD + "Failed to merge" + clr.RESET)
        print("Reason:", e)
    


def parse_args():
    parser = argparse.ArgumentParser()  # top level parser
    subparsers = parser.add_subparsers(help="The mode to run in")

    ## Scan parser
    scan_parser = subparsers.add_parser(
        "scan", help="Scan mode. Scan target(s) using the pre-defined Nmap scans"
    )
    scan_parser.add_argument(
        "target",
        help="Target(s) to scan. Can be in the format of a single IP or CIDR (192.168.0.0/24 for example)",
    )
    scan_parser.add_argument(
        "database",
        help="Database to store results into. Will create one if it does not exist",
    )
    scan_parser.add_argument(
        "-o",
        "--output",
        help="Output directory. Will write the nmap files into this dir, divided in subdirs.",
    )
    scan_parser.add_argument(
        "-i",
        "--version-intensity",
        help="Version intensity. Overwrites the default value of 6",
    )
    scan_parser.add_argument(
        "-t",
        "--threads",
        help="Maximum number of threads to run. Default is 1",
        type=int,
    )
    scan_parser.add_argument(
        "-nfp",
        "--no-full-portscan",
        help="Do no do a full portscan for all hosts. This could save a lot of time.",
        action="store_true",
    )
    scan_parser.set_defaults(func=scan)

    autoscan_parser = subparsers.add_parser(
        "autoscan",
        help="Autoscan mode. Autoscan local network using the pre-defined Nmap scans",
    )
    autoscan_parser.add_argument(
        "cidr",
        help="Subnet range to scan. Will result in scanning a network like 192.168.1.0/24 or 192.168.0.0/16 for example",
        choices=["16", "24"],
    )
    autoscan_parser.add_argument(
        "database",
        help="Database to store results into. Will create one if it does not exist",
    )
    autoscan_parser.add_argument(
        "-o",
        "--output",
        help="Output directory. Will write the nmap files into this dir, divided in subdirs.",
    )
    autoscan_parser.add_argument(
        "-i",
        "--version-intensity",
        help="Version intensity. Overwrites the default value of 6",
    )
    autoscan_parser.add_argument(
        "-t",
        "--threads",
        help="Maximum number of threads to run. Default is 1",
        type=int,
    )
    autoscan_parser.set_defaults(func=autoscan)

    ## Read parser
    read_parser = subparsers.add_parser(
        "read", help="Read mode. Read information on target(s), found during scan mode"
    )
    read_parser.add_argument(
        "-t",
        "--target",
        help="Target(s) to read information of. Quick way to get target information.",
    )
    read_parser.add_argument(
        "-s",
        "--service",
        help="List all targets with the specified service running, along with the port numbers",
    )
    read_parser.add_argument("database", help="Database to read from.")
    read_parser.set_defaults(func=read)

    # inter_parser = subparsers.add_parser('interactive', help='Interactive mode. Interactively request information from the database.')
    # inter_parser.set_defaults(func=interactive)
    xml_read_parser = subparsers.add_parser(
        "xml", help="Read information from a Nmap XML file"
    )
    xml_read_parser.add_argument("file", help="Nmap XML file to read")
    xml_read_parser.add_argument(
        "-t", "--target", help="Read results for specific target"
    )
    xml_read_parser.add_argument(
        "-s", "--service", help="Show only hosts with a specific service running"
    )
    xml_read_parser.add_argument(
        "-p", "--port", help="Show only hosts with a specific port opened"
    )
    xml_read_parser.add_argument(
        "-os", "--os-type", help="Show only hosts with a specific Operating System type"
    )
    xml_read_parser.add_argument(
        "-pr",
        "--product",
        help='Show only hosts with a specific product, such as "OpenSSH"',
    )
    xml_read_parser.set_defaults(func=xml)

    merge_parser = subparsers.add_parser(
        "merge", help="Merge multiple Nmap XML files into one"
    )
    merge_parser.add_argument(
        "files", help="Nmap XML files to merge", type=str, nargs="+"
    )
    merge_parser.add_argument(
        "-o",
        "--new-file",
        help="The file to write the results into",
        default="new.xml",
        const="newfile.xml",
        nargs="?",
        type=str,
    )
    merge_parser.set_defaults(func=merge)

    args = parser.parse_args()
    return args, parser


def main():
    version="0.1.4"
    print(INIT, "Starting Nwrapy", version)

    args, parser = parse_args()

    try:
        args.func(args)
    except AttributeError:
        parser.print_help()


if __name__ == "__main__":
    main()



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    readme=f.read()

setup(name='nwrapy',
      version='0.1.4',
      description='Nmap wrapper',
      author='Jochem Stevense',
      author_email='',
      #long_description=readme,
      url='https://github.com/SneakyBeagle/nwrapy',
      project_urls={
          "Bug Tracker": "https://github.com/SneakyBeagle/nwrapy/issues"
      },
      classifiers=[
          "Programming Language :: Python :: 3",
      ],
      packages=find_packages(include=['nwrapy']),
      license='GPLv3',
      python_requires=">=3.6",
      entry_points={
          'console_scripts': ['nwrapy=nwrapy:main']
          }
      )




class colours():
    def __init__(self):
        self.RESET = '\033[0m'
        self.BOLD = '\033[1m'
        self.DIM = '\033[2m'
        self.ITALIC = '\033[3m'
        self.UNDERLINE = '\033[4m'
        self.BLINK = '\033[5m'
        self.BLINK2 = '\033[6m'
        self.BACKGROUND = '\033[7m'
        self.INVISIBLE = '\033[8m'
        self.CROSSED = '\033[9m'
        
        self.DOUBLEUNDERLINE='\033[21m'
        
        self.BOLD_WHITE = '\033[1m\033[37m'
        self.FAINT_WHITE = '\033[2m\033[37m'
        
        self.BLACK = '\033[30m'
        self.RED = '\033[31m'
        self.GREEN = '\033[32m'
        self.YELLOW = '\033[33m'
        self.BLUE = '\033[34m'
        self.PURPLE = '\033[35m'
        self.LIGHTBLUE = '\033[36m'
        self.GREY = '\033[37m'

    def help(self):
        print(self.__dict__)

    def get_all(self):
        return self.__dict__

    def test2(self):
        for k,v in self.get_all().items():
            print(k, v, v.split('\0'), self.RESET)
    
    def test(self):
        #print(dir(self))
        # Styles
        for i in range(0,10):
            print('\033['+str(i)+'m'+'\\033['+str(i)+'m'+self.RESET)

        # Colours
        for i in range(0,10):
            print('\033[3'+str(i)+'m'+'\\033[3'+str(i)+'m'+self.RESET)

        
        for i in range(0,10):
            print('\033[2'+str(i)+'m'+'\\033[2'+str(i)+'m'+self.RESET)
        

clr=colours()

if __name__=='__main__':
    clr.help()
    clr.test()
    clr.test2()
